from ensembl_config import ensembl_http_status_codes

"""
	Implements custom exceptions for the EnsEMBL REST API
"""

class EnsemblRestError(Exception):
	"""
		Generic error class, catch-all for most EnsemblRest issues.
		Special cases are handled by EnsemblRestRateLimitError and EnsemblRestServiceUnavailable.
	"""
	def __init__(self, msg, error_code=None, rate_reset=None, rate_limit=None, rate_remaining=None):
		self.error_code = error_code

		if error_code is not None and error_code in ensembl_http_status_codes:
			msg = 'EnsEMBL REST API returned a %s (%s), %s' % \
            		(error_code, ensembl_http_status_codes[error_code][0], msg)

		super(EnsemblRestError, self).__init__(msg)

	@property
	def msg(self):
		return self.args[0]

class EnsemblRestRateLimitError(EnsemblRestError):
	"""
		Raised when you've hit a rate limit.
		The amount of seconds to retry your request in will be appended to the message.
	"""
	def __init__(self, msg, error_code, rate_reset=None, rate_limit=None, rate_remaining=None):
		if isinstance(rate_limit, int):
			msg = '%s (Rate limit hit:  %d seconds)' % (msg, rate_reset, rate_limit, rate_remaining)
		EnsemblRestError.__init__(self, msg, error_code=error_code)

class EnsemblRestServiceUnavailable(EnsemblRestError):
	"""
		Raised when the service is down.
	"""	
	pass


"""
	Configuration information for the EnsEMBL REST API
"""

ensembl_default_url = 'http://beta.rest.ensembl.org'
ensembl_genomes_url = 'http://test.rest.ensemblgenomes.org'

ensembl_api_table = {
	# Comparative Genomics
	'getGeneTreeById': {
		'url': '/genetree/id/{{id}}',
		'method': 'GET',
		'content_type': 'text/x-nh'
	},
	'getGeneTreeByMemberId': {
		'url': '/genetree/member/id/{{id}}',
		'method': 'GET',
		'content_type': 'text/x-phyloxml+xml'
	},
	'getGeneTreeByMemberSymbol': {
		'url': '/genetree/member/symbol/{{species}}/{{symbol}}',
		'method': 'GET',
		'content_type': 'text/x-phyloxml+xml'
	},
	'getHomologyById': {
		'url': '/homology/id/{{id}}',
		'method': 'GET',
		'content_type': 'application/json'
	},
	'getHomologyBySymbol': {
		'url': '/homology/symbol/{{species}}/{{symbol}}',
		'method': 'GET',
		'content_type': 'application/json'
	},

	# Cross References
	'getXrefsById': {
		'url': '/xrefs/id/{{id}}',
		'method': 'GET',
		'content_type': 'application/json'
	},
	'getXrefsByName': {
		'url': '/xrefs/name/{{species}}/{{name}}',
		'method': 'GET',
		'content_type': 'application/json'
	},
	'getXrefsBySymbol': {
		'url': '/xrefs/symbol/{{species}}/{{symbol}}',
		'method': 'GET',
		'content_type': 'application/json'
	},

	# Features
	'getFeatureById': {
		'url': '/feature/id/{{id}}?feature={{feature}}',
		'method': 'GET',
		'content_type': 'application/json'
	},
	'getFeatureByRegion': {
		'url': '/feature/region/{{species}}/{{region}}?feature={{feature}}',
		'method': 'GET',
		'content_type': 'application/json'
	},

	# Information
	'getAssemblyInfo': {
		'url': '/assembly/info/{{species}}',
		'method': 'GET',
		'content_type': 'application/json'
	},
	'getAssemblyInfoRegion': {
		'url': '/assembly/info/{{species}}/{{region_name}}',
		'method': 'GET',
		'content_type': 'application/json'
	},
	'getInfoComparas': {
		'url': '/info/comparas',
		'method': 'GET',
		'content_type': 'application/json'
	},
	'getInfoData': {
		'url': '/info/data',
		'method': 'GET',
		'content_type': 'application/json'
	},
	'getInfoPing': {
		'url': '/info/ping',
		'method': 'GET',
		'content_type': 'application/json'
	},
	'getInfoRest': {
		'url': '/info/rest',
		'method': 'GET',
		'content_type': 'application/json'
	},
	'getInfoSoftware': {
		'url': '/info/software',
		'method': 'GET',
		'content_type': 'application/json'
	},
	'getInfoSpecies': {
		'url': '/info/species',
		'method': 'GET',
		'content_type': 'application/json'
	},

	# Lookup
	'getLookupById': {
		'url': '/lookup/id/{{id}}',
		'method': 'GET',
		'content_type': 'application/json'
	},

	# Mapping
	'getMapAssemblyOneToTwo': {
		'url': '/map/{{species}}/{{asm_one}}/{{region}}/{{asm_two}}',
		'method': 'GET',
		'content_type': 'application/json'
	},
	'getMapCdnaToRegion': {
		'url': '/map/cdna/{{id}}/{{region}}',
		'method': 'GET',
		'content_type': 'application/json'
	},
	'getMapCdsToRegion': {
		'url': '/map/cds/{{id}}/{{region}}',
		'method': 'GET',
		'content_type': 'application/json'
	},
	'getMapTranslationToRegion': {
		'url': '/map/translation/{{id}}/{{region}}',
		'method': 'GET',
		'content_type': 'application/json'
	},

	# Ontologies and Taxonomy
	'getAncestorsById': {
		'url': '/ontology/ancestors/{{id}}',
		'method': 'GET',
		'content_type': 'application/json'
	},
	'getAncestorsChartById': {
		'url': '/ontology/ancestors/chart/{{id}}',
		'method': 'GET',
		'content_type': 'application/json'
	},
	'getDescendentsById': {
		'url': '/ontology/descendents/{{id}}',
		'method': 'GET',
		'content_type': 'application/json'
	},
	'getOntologyById': {
		'url': '/ontology/id/{{id}}',
		'method': 'GET',
		'content_type': 'application/json'
	},
	'getOntologyByName': {
		'url': '/ontology/name/{{name}}',
		'method': 'GET',
		'content_type': 'application/json'
	},
	'getTaxonomyClassificationById': {
		'url': '/taxonomy/classification/{{id}}',
		'method': 'GET',
		'content_type': 'application/json'
	},
	'getTaxonomyById': {
		'url': '/taxonomy/id/{{id}}',
		'method': 'GET',
		'content_type': 'application/json'
	},

	# Sequences
	'getSequenceById': {
		'url': '/sequence/id/{{id}}',
		'method': 'GET',
		'content_type': 'application/json'
	},
	'getSequenceByRegion': {
		'url': '/sequence/region/{{species}}/{{region}}',
		'method': 'GET',
		'content_type': 'application/json'
	},

	# Variation
	'getVariantConsequencesByRegion': {
		'url': '/vep/{{species}}/{{region}}/{{allele}}/consequences',
		'method': 'GET',
		'content_type': 'application/json'
	},
	'getVariantConsequencesById': {
		'url': '/vep/{{species}}/id/{{id}}/consequences',
		'method': 'GET',
		'content_type': 'application/json'
	},
}

ensembl_http_status_codes = {
	200: ('OK', 'Request was a success. Only process data from the service when you receive this code'),
	400: ('Bad Request', 'Occurs during exceptional circumstances such as the service is unable to find an ID. Check if the response Content-type was JSON. If so the JSON object is an exception hash with the message keyed under error'),
	404: ('Not Found', 'Indicates a badly formatted request. Check your URL'),
	415: ('Unsupported Media Type', 'The server is refusing to service the request because the entity of the request is in a format not supported by the requested resource for the requested method'),
	429: ('Too Many Requests', 'You have been rate-limited; wait and retry. The headers X-RateLimit-Reset, X-RateLimit-Limit and X-RateLimit-Remaining will inform you of how long you have until your limit is reset and what that limit was. If you get this response and have not exceeded your limit then check if you have made too many requests per second.'),
	503: ('Service Unavailable', 'The service is temporarily down; retry after a pause'),
}


"""
	EnsemblRest is a library for Python that wrap the EnsEMBL REST API.
	It simplifies all the API endpoints by abstracting them away from the
	end user and thus also ensures that an amendments to the library and/or
	EnsEMBL REST API won't cause major problems to the end user.

	Thanks to Ryan McGrath's <ryan@venodesigns.net> Twython API for
	assisting with designing this code <https://github.com/ryanmcgrath/twython>.

	Any questions, comments or issues can be addressed to gawbul@gmail.com.
"""

__author__ = "Steve Moss"
__copyright__ = "Copyright 2013, Steve Moss"
__credits__ = ["Steve Moss"]
__license__ = "GPL"
__version__ = "0.1.7b"
__maintainer__ = "Steve Moss"
__email__ = "gawbul@gmail.com"
__status__ = "beta"

import re
import sys
import requests
import json

from .ensembl_config import ensembl_default_url, ensembl_genomes_url, ensembl_api_table
from .exceptions import EnsemblRestError, EnsemblRestRateLimitError, EnsemblRestServiceUnavailable

class EnsemblRest(object):
	""" The REST API object """
	def __init__(self, server=None, proxies=None):
		# set REST API server url
		if not server == None:
			if server == "default":
				self.base_url = ensembl_default_url
			elif server == "genomes":
				self.base_url = ensembl_genomes_url
			else:
				self.base_url = server
		else:
			self.base_url = ensembl_default_url

		# setup requests session
		self.client = requests.Session()
		self.client.headers = {'User-Agent': 'pyEnsemblRest v' + __version__}
		self.client.proxies = proxies

        # register available functions to allow listing name when debugging
		def regFunc(key):
			return lambda **kwargs: self._constructFunc(key, **kwargs)

		# iterate over ensembl_api_table keys and add key to class namespace
		for key in ensembl_api_table.keys():
			self.__dict__[key] = regFunc(key)

	def _constructFunc(self, api_call, **kwargs):
		""" Function constructor """
		# Go through and replace any moustaches that are in the API url.
		fn = ensembl_api_table[api_call]
		url = re.sub('\{\{(?P<m>[a-zA-Z_]+)\}\}',
						lambda m: "%s" % kwargs.get(m.group(1)), self.base_url + fn['url'])
		content = self._request(url, method=fn['method'], content_type=fn['content_type'], params=kwargs)

		return content

	def _request(self, url, method='GET', content_type='application/json', params=None, files=None, api_call=None):
		""" Internal response generator, no sense in repeating the same code twice, right? ;) """
		method = method.lower()
		if not method in ('get', 'post'):
			raise Exception('Method must be either GET or POST')

		# set content type
		if not content_type:
			raise Exception('Content-Type must be provided')
		self.client.headers['Content-Type'] = content_type

		params = params or {}
		# convert params.items to unicode to be nice to requests
		for k, v in params.items():
			if isinstance(v, (int, bool)):
				params[k] = u'%s' % v

		func = getattr(self.client, method)
		if method == 'get':
			response = func(url, params=params)
		else:
			response = func(url, data=params, files=files)
		content = response.content.decode('utf-8')

		# store last call to api for debugging purposes
		self._last_call = {
			'api_call': api_call,
			'api_error': None,
			'cookies': response.cookies,
			'headers': response.headers,
			'status_code': response.status_code,
			'url': response.url,
			'content': content,
		}
		
		# wrap the json loads in a try, and defer an error
        # why? EnsEMBL REST API will return invalid json with an error code in the headers
		json_error = False
		if content_type == 'application/json':
			try:
				try:
					# try to get json
					content = content.json()
				except AttributeError:
					# if unicode detected
					content = json.loads(content)
			except ValueError:
				json_error = True
				content = {}
		
		if response.status_code > 304:
			# If there is no error message, use a default.
			error_message = content.get('error', 'An error occurred processing your request.')
			self._last_call['api_error'] = error_message

			ExceptionType = EnsemblRestError
			if response.status_code == 429:
				# EnsEMBL REST API always returns 429 when rate limit is exceeded
				ExceptionType = EnsemblRestRateLimitError

			raise ExceptionType(error_message,
								error_code=response.status_code,
								rate_reset=response.headers.get('X-RateLimit-Reset'),
								rate_limit=response.headers.get('X-RateLimit-Limit'),
								rate_remaining=response.headers.get('X-RateLimit-Remaining')
								)

		# if we have a json error here, then it's not an official EnsEMBL REST API error
		if json_error and not response.status_code in (200, 201, 202):
			raise EnsemblRestError('Response was not valid JSON, unable to decode.')

		return content


"""
	EnsemblRest is a library for Python that wrap the EnsEMBL REST API.
	It simplifies all the API endpoints by abstracting them away from the
	end user and thus also ensures that an amendments to the library and/or
	EnsEMBL REST API won't cause major problems to the end user.

	Thanks to Ryan McGrath's <ryan@venodesigns.net> Twython API for
	assisting with designing this code <https://github.com/ryanmcgrath/twython>.

	Any questions, comments or issues can be addressed to gawbul@gmail.com.
"""

__author__ = "Steve Moss"
__version__ = "0.1.7b"

from .ensemblrest import EnsemblRest
from .exceptions import EnsemblRestError, EnsemblRestRateLimitError, EnsemblRestServiceUnavailable


from setuptools import setup
from setuptools import find_packages

__author__ = 'Steve Moss'
__email__ = 'gawbul@gmail.com'
__version__ = '0.1.7b'

setup(
    # Basic package information.
    name='pyensemblrest',
    version=__version__,
    packages=find_packages(),

    # Packaging options.
    include_package_data=True,

    # Package dependencies.
    install_requires=['requests>=1.0.0, <2.0.0'],

    # Metadata for PyPI.
    author='Steve Moss',
    author_email='gawbul@gmail.com',
    license='GPLv3',
    url='http://github.com/pyopensci/pyensemblrest/tree/master',
    keywords='ensembl python rest api',
    description='An easy way to access EnsEMBL data with Python.',
    long_description=open('README.rst').read(),
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Internet',
    ]
)